/* Copyright (C) 2009 Kentoku Shiba

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifdef USE_PRAGMA_INTERFACE
#pragma interface
#endif

#define VP_TABLE_INFO_MAX_LEN 64
#define VP_MAX_KEY_LENGTH 16384

class ha_vp: public handler
{
public:
  THR_LOCK_DATA      lock;
  VP_SHARE           *share;
  String             *blob_buff;

  uint               sql_command;
  int                lock_mode;
  enum thr_lock_type lock_type_sto;
  int                lock_type_ext;

  TABLE_LIST         *part_tables;
  uint               table_lock_count;
  uchar              *use_tables;
  uchar              *use_tables2;
  uchar              *use_tables3;
  uchar              *select_ignore;
  uchar              *select_ignore_with_lock;
  uchar              *update_ignore;
  uchar              *work_bitmap;
  uchar              *work_bitmap2;
  uchar              *work_bitmap3;
  uchar              *work_bitmap4;
  bool               child_keyread;
  bool               single_table;
  bool               update_request;
  bool               set_used_table;
  bool               bulk_insert;
  bool               init_bitmap;
  int                child_table_idx;
  int                child_key_idx;
  uchar              *child_key;
  uchar              child_key_different[MAX_KEY_LENGTH];
  uchar              child_end_key_different[MAX_KEY_LENGTH];
  key_range          child_start_key;
  key_range          child_end_key;
  KEY_MULTI_RANGE    *child_found_range;
  int                dup_table_idx;
#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
  Field              ***top_table_field_for_childs;
  uint               allocated_top_table_fields;
  bool               top_table_self;
#endif
#ifndef WITHOUT_VP_BG_ACCESS
  VP_BG_BASE         *bg_base;
#endif
  longlong           additional_table_flags;

  ha_vp();
  ha_vp(
    handlerton *hton,
    TABLE_SHARE *table_arg
  );
  const char **bas_ext() const;
  int open(
    const char* name,
    int mode,
    uint test_if_locked
  );
  int close();
  int external_lock(
    THD *thd,
    int lock_type
  );
  uint lock_count() const;
  THR_LOCK_DATA **store_lock(
    THD *thd,
    THR_LOCK_DATA **to,
    enum thr_lock_type lock_type
  );
  int reset();
  int extra(
    enum ha_extra_function operation
  );
  int extra_opt(
    enum ha_extra_function operation,
    ulong cachesize
  );
  int index_init(
    uint idx,
    bool sorted
  );
  int index_end();
  int index_read_map(
    uchar *buf,
    const uchar *key,
    key_part_map keypart_map,
    enum ha_rkey_function find_flag
  );
  int index_read_last_map(
    uchar *buf,
    const uchar *key,
    key_part_map keypart_map
  );
  int index_next(
    uchar *buf
  );
  int index_prev(
    uchar *buf
  );
  int index_first(
    uchar *buf
  );
  int index_last(
    uchar *buf
  );
  int index_next_same(
    uchar *buf,
    const uchar *key,
    uint keylen
  );
  int read_range_first(
    const key_range *start_key,
    const key_range *end_key,
    bool eq_range,
    bool sorted
  );
  int read_range_next();
  int read_multi_range_first(
    KEY_MULTI_RANGE **found_range_p,
    KEY_MULTI_RANGE *ranges,
    uint range_count,
    bool sorted,
    HANDLER_BUFFER *buffer
  );
  int read_multi_range_next(
    KEY_MULTI_RANGE **found_range_p
  );
  int rnd_init(
    bool scan
  );
  int rnd_end();
  int rnd_next(
    uchar *buf
  );
  void position(
    const uchar *record
  );
  int rnd_pos(
    uchar *buf,
    uchar *pos
  );
  int info(
    uint flag
  );
  ha_rows records_in_range(
    uint idx,
    key_range *start_key,
    key_range *end_key
  );
  const char *table_type() const;
  ulonglong table_flags() const;
  const char *index_type(
    uint key_number
  );
  ulong index_flags(
    uint idx,
    uint part,
    bool all_parts
  ) const;
  uint max_supported_record_length() const;
  uint max_supported_keys() const;
  uint max_supported_key_parts() const;
  uint max_supported_key_length() const;
  uint max_supported_key_part_length() const;
  uint8 table_cache_type();
  int update_auto_increment();
  void set_next_insert_id(
    ulonglong id
  );
  void get_auto_increment(
    ulonglong offset,
    ulonglong increment,
    ulonglong nb_desired_values,
    ulonglong *first_value,
    ulonglong *nb_reserved_values
  );
  void restore_auto_increment(
    ulonglong prev_insert_id
  );
  void release_auto_increment();
  int reset_auto_increment(
    ulonglong value
  );
  void start_bulk_insert(
    ha_rows rows
  );
  int end_bulk_insert();
  int write_row(
    uchar *buf
  );
  int update_row(
    const uchar *old_data,
    uchar *new_data
  );
  int delete_row(
    const uchar *buf
  );
  int delete_all_rows();
  double scan_time();
  double read_time(
    uint index,
    uint ranges,
    ha_rows rows
  );
  const key_map *keys_to_use_for_scanning();
  ha_rows estimate_rows_upper_bound();
  bool get_error_message(
    int error,
    String *buf
  );
  int create(
    const char *name,
    TABLE *form,
    HA_CREATE_INFO *info
  );
  void update_create_info(
    HA_CREATE_INFO* create_info
  );
  int rename_table(
    const char *from,
    const char *to
  );
  int delete_table(
    const char *name
  );
  bool is_crashed() const;
  bool auto_repair() const;
  int disable_indexes(
    uint mode
  );
  int enable_indexes(
    uint mode
  );
  int check(
    THD* thd,
    HA_CHECK_OPT* check_opt
  );
  int repair(
    THD* thd,
    HA_CHECK_OPT* check_opt
  );
  bool check_and_repair(
    THD *thd
  );
  int analyze(
    THD* thd,
    HA_CHECK_OPT* check_opt
  );
  int optimize(
    THD* thd,
    HA_CHECK_OPT* check_opt
  );
#ifdef HANDLER_HAS_TOP_TABLE_FIELDS
  int set_top_table_and_fields(
    TABLE *top_table,
    Field **top_table_field,
    uint top_table_fields,
    bool self
  );
  int set_top_table_and_fields(
    TABLE *top_table,
    Field **top_table_field,
    uint top_table_fields
  );
  const COND *cond_push(
    const COND *cond
  );
  void cond_pop();
#endif

private:
  int choose_child_index(
    uint idx,
    uchar *read_set,
    uchar *write_set,
    int *table_idx,
    int *key_idx
  );
  int choose_child_tables(
    uchar *read_set,
    uchar *write_set
  );
  void clear_child_bitmap(
    int table_idx
  );
  uchar *create_child_key(
    const uchar *key_same,
    uchar *key_different,
    key_part_map keypart_map
  );
  int get_child_record_by_idx(
    int table_idx,
    my_ptrdiff_t ptr_diff
  );
  int get_child_record_by_pk(
    my_ptrdiff_t ptr_diff
  );
  bool set_child_bitmap(
    uchar *bitmap,
    int table_idx,
    bool write_flg
  );
  void set_child_record_for_update(
    my_ptrdiff_t ptr_diff,
    int record_idx,
    bool write_flg,
    bool use_table_chk
  );
  int search_by_pk(
    int table_idx,
    int record_idx,
    VP_KEY_COPY *key_copy,
    my_ptrdiff_t ptr_diff,
    uchar **table_key
  );
  int search_by_pk_for_update(
    int table_idx,
    int record_idx,
    VP_KEY_COPY *vp_key_copy,
    my_ptrdiff_t ptr_diff,
    int bgu_mode
  );
  int create_bg_thread(
    VP_BG_BASE *base
  );
  void free_bg_thread(
    VP_BG_BASE *base
  );
  void bg_kick(
    VP_BG_BASE *base
  );
  void bg_wait(
    VP_BG_BASE *base
  );
};
